from pathlib import Path
from typing import Union

from polylith import configuration, repo


def create(name: Union[str, None], description: Union[str, None], fn):
    root = repo.get_workspace_root(Path.cwd())
    namespace = configuration.get_namespace_from_config(root)

    if not name:
        raise ValueError("Please add a name by using --name")

    if not namespace:
        raise ValueError(
            "Didn't find a namespace. Expected to find it under [tool.polylith] in workspace.toml or pyproject.toml."
        )

    options = {
        "namespace": namespace,
        "package": name,
        "description": description,
        "modulename": "core",
    }
    fn(root, options)



from functools import partial
from pathlib import Path

from cleo.helpers import option
from poetry.console.commands.command import Command
from polylith import commands, configuration, info, repo
from polylith.poetry.commands.check import command_options
from polylith.poetry.internals import (
    distributions,
    filter_projects_data,
    find_third_party_libs,
)


class LibsCommand(Command):
    name = "poly libs"
    description = "Show third-party libraries used in the workspace."

    options = command_options + [
        option(
            long_name="short",
            short_name="s",
            description="Print short view",
            flag=True,
        ),
    ]

    def merged_project_data(self, data: dict) -> dict:
        name = data["name"]
        path = data["path"]

        try:
            third_party_libs = find_third_party_libs(self.poetry, path)
            return {
                **data,
                **{"deps": {"items": third_party_libs, "source": "poetry.lock"}},
            }
        except ValueError as e:
            self.line_error(f"{name}: <error>{e}</error>")
            return data

    def handle(self) -> int:
        root = repo.get_workspace_root(Path.cwd())
        dists_fn = partial(distributions, None, root)

        options = {
            "strict": self.option("strict"),
            "alias": self.option("alias"),
            "short": self.option("short"),
            "dists_fn": dists_fn,
        }

        directory = self.option("directory")
        ns = configuration.get_namespace_from_config(root)

        all_projects_data = info.get_projects_data(root, ns)
        projects_data = filter_projects_data(self.poetry, directory, all_projects_data)

        merged_projects_data = [
            self.merged_project_data(data) for data in projects_data
        ]

        results = commands.libs.run(root, ns, merged_projects_data, options)
        commands.libs.run_library_versions(projects_data, all_projects_data, options)

        return 0 if all(results) else 1



import difflib
from operator import itemgetter
from pathlib import Path
from typing import List, Set, Union

from polylith import workspace
from polylith.libs import grouping
from polylith.reporting import theme
from rich import box, markup
from rich.console import Console
from rich.padding import Padding
from rich.table import Table


def get_third_party_imports(root: Path, ns: str, project_data: dict) -> dict:
    bases = {b for b in project_data.get("bases", [])}
    components = {c for c in project_data.get("components", [])}

    bases_paths = workspace.paths.collect_bases_paths(root, ns, bases)
    components_paths = workspace.paths.collect_components_paths(root, ns, components)

    bases_imports = grouping.get_third_party_imports(root, bases_paths)
    components_imports = grouping.get_third_party_imports(root, components_paths)

    return {"bases": bases_imports, "components": components_imports}


def flatten_imports(brick_imports: dict, brick: str) -> Set[str]:
    return set().union(*brick_imports.get(brick, {}).values())


def flatten_brick_imports(brick_imports: dict) -> Set[str]:
    bases_imports = flatten_imports(brick_imports, "bases")
    components_imports = flatten_imports(brick_imports, "components")

    return set().union(bases_imports, components_imports)


def filter_close_matches(
    unknown_imports: Set[str], dependencies: Set[str], cutoff: float
) -> Set[str]:
    unknowns = {str.lower(u) for u in unknown_imports}
    deps = {str.lower(d).replace("-", "_") for d in dependencies}

    return {
        u for u in unknowns if not difflib.get_close_matches(u, deps, cutoff=cutoff)
    }


def get_unknowns(brick_imports: dict, deps: Set[str]) -> Set[str]:
    imports = flatten_brick_imports(brick_imports)

    return imports.difference(deps)


def calculate_diff(
    brick_imports: dict, deps: Set[str], is_strict: bool = False
) -> Set[str]:
    unknown_imports = get_unknowns(brick_imports, deps)

    cutoff = 0.6 if not is_strict else 1

    return filter_close_matches(unknown_imports, deps, cutoff)


def print_libs_summary() -> None:
    console = Console(theme=theme.poly_theme)

    console.print(Padding("[data]Libraries in bricks[/]", (1, 0, 0, 0)))


def print_libs_in_bricks(brick_imports: dict) -> None:
    bases_imports = flatten_imports(brick_imports, "bases")
    components_imports = flatten_imports(brick_imports, "components")

    if not bases_imports and not components_imports:
        return

    console = Console(theme=theme.poly_theme)
    table = Table(box=box.SIMPLE_HEAD)

    bases = brick_imports.get("bases", {})
    components = brick_imports.get("components", {})

    table.add_column("[data]brick[/]")
    table.add_column("[data]library[/]")

    for brick, imports in sorted(components.items(), key=itemgetter(0)):
        table.add_row(f"[comp]{brick}[/]", ", ".join(sorted(imports)))

    for brick, imports in sorted(bases.items(), key=itemgetter(0)):
        table.add_row(f"[base]{brick}[/]", ", ".join(sorted(imports)))

    console.print(table, overflow="ellipsis")


def print_missing_installed_libs(
    brick_imports: dict,
    third_party_libs: Set[str],
    project_name: str,
    is_strict: bool = False,
) -> bool:
    diff = calculate_diff(brick_imports, third_party_libs, is_strict)

    if not diff:
        return True

    console = Console(theme=theme.poly_theme)

    missing = ", ".join(sorted(diff))

    console.print(
        f"[data]Could not locate all libraries in [/][proj]{project_name}[/]. [data]Caused by missing dependencies?[/]"
    )

    console.print(f":thinking_face: {missing}")
    return False


def printable_version(version: Union[str, None], is_same_version: bool) -> str:
    ver = version or "-"
    markup = "data" if is_same_version else "bold"

    return f"[{markup}]{ver}[/]"


def get_version(lib: str, project_data: dict) -> str:
    return project_data["deps"]["items"].get(lib)


def find_version(
    lib: str, project_name: str, projects_data: List[dict]
) -> Union[str, None]:
    project_data = next(p for p in projects_data if p["name"] == project_name)

    return get_version(lib, project_data)


def printable_header(header: str, short: bool) -> str:
    return "\n".join(header) if short else header


def is_same_version(versions: list) -> bool:
    unique = set([v for v in versions if v])

    return len(unique) == 1 if unique else True


def libs_in_projects_table(
    development_data: dict,
    projects_data: List[dict],
    libraries: set,
    options: dict,
) -> Table:
    table = Table(box=box.SIMPLE_HEAD)

    short = options["short"]

    project_names = sorted({p["name"] for p in projects_data})
    project_headers = [f"[proj]{printable_header(n, short)}[/]" for n in project_names]
    dev_header = printable_header("development", short)
    headers = ["[data]library[/]"] + project_headers + [f"[data]{dev_header}[/]"]

    for header in headers:
        table.add_column(header)

    for lib in sorted(libraries):
        proj_versions = [find_version(lib, n, projects_data) for n in project_names]
        dev_version = get_version(lib, development_data)

        is_same = is_same_version(proj_versions + [dev_version])
        printable_proj_versions = [printable_version(v, is_same) for v in proj_versions]
        printable_dev_version = printable_version(dev_version, is_same)

        cols = [markup.escape(lib)] + printable_proj_versions + [printable_dev_version]

        table.add_row(*cols)

    return table


def flattened_lib_names(projects_data: List[dict]) -> Set[str]:
    return {k for proj in projects_data for k, _v in proj["deps"]["items"].items()}


def print_libs_in_projects(
    development_data: dict, projects_data: List[dict], options: dict
) -> None:
    flattened = flattened_lib_names(projects_data)

    if not flattened:
        return

    table = libs_in_projects_table(development_data, projects_data, flattened, options)

    console = Console(theme=theme.poly_theme)

    console.print(Padding("[data]Library versions in projects[/]", (1, 0, 0, 0)))
    console.print(table, overflow="ellipsis")


def has_different_version(
    lib: str, development_data: dict, projects_data: List[dict]
) -> bool:
    proj_versions = [get_version(lib, p) for p in projects_data]
    dev_version = get_version(lib, development_data)

    return not is_same_version(proj_versions + [dev_version])


def libs_with_different_versions(
    development_data: dict, projects_data: List[dict]
) -> Set[str]:
    flattened = flattened_lib_names(projects_data)
    return {
        f
        for f in flattened
        if has_different_version(f, development_data, projects_data)
    }


def print_libs_with_different_versions(
    libraries: Set[str],
    development_data: dict,
    projects_data: List[dict],
    options: dict,
) -> None:
    if not libraries:
        return

    table = libs_in_projects_table(development_data, projects_data, libraries, options)

    console = Console(theme=theme.poly_theme)
    console.print(
        Padding("[data]Different library versions in projects[/]", (1, 0, 0, 0))
    )
    console.print(table, overflow="ellipsis")



from pathlib import Path

from polylith import repo

workspace_template = """\
# A Python Polylith repo

## Docs
The official Polylith documentation:
[high-level documentation](https://polylith.gitbook.io/polylith)

A Python implementation of the Polylith tool:
[python-polylith](https://github.com/DavidVujic/python-polylith)
"""

brick_template = """\
# {name} {brick}

{description}
"""


def create_readme(path: Path, template: str, **kwargs) -> None:
    fullpath = path / repo.readme_file

    if fullpath.exists():
        return

    with fullpath.open("w", encoding="utf-8") as f:
        f.write(template.format(**kwargs))


def create_workspace_readme(path: Path, namespace: str) -> None:
    create_readme(path, workspace_template, namespace=namespace)


def create_brick_readme(path: Path, options: dict) -> None:
    brick = options["brick"]
    package = options["package"]
    description = options["description"]

    b = "component" if brick in repo.components_dir else "base"

    create_readme(
        path, brick_template, name=package, brick=b, description=description or ""
    )



from pathlib import Path
from typing import List, Set

from polylith import check, distributions, libs


def with_third_party_libs_from_lock_file(project_data: dict) -> dict:
    lock_file_data = libs.pick_lock_file(project_data)

    if not lock_file_data:
        return project_data

    filename = lock_file_data["filename"]
    filetype = lock_file_data["filetype"]

    third_party_libs = libs.extract_libs(project_data, filename, filetype)

    merged = {
        **project_data,
        **{"deps": {"items": third_party_libs, "source": filename}},
    }

    return merged


def collect_known_aliases(project_data: dict, options: dict) -> Set[str]:
    deps = project_data["deps"]
    library_alias = options["alias"]

    return distributions.known_aliases_and_sub_dependencies(
        deps, library_alias, options
    )


def check_libs_versions(
    projects_data: List[dict], all_projects_data: List[dict], options: dict
) -> bool:
    is_strict = options["strict"]
    is_quiet = options["quiet"]

    if not is_strict:
        return True

    development_data = next(p for p in all_projects_data if p["type"] == "development")

    libraries = libs.report.libs_with_different_versions(
        development_data, projects_data
    )

    if not is_quiet:
        libs.report.print_libs_with_different_versions(
            libraries, development_data, projects_data, options
        )

    return False if libraries else True


def run(root: Path, ns: str, project_data: dict, options: dict) -> bool:
    is_verbose = options["verbose"]
    is_quiet = options["quiet"]
    is_strict = options["strict"]

    name = project_data["name"]

    collected_imports = check.report.collect_all_imports(root, ns, project_data)
    collected_libs = collect_known_aliases(project_data, options)

    details = check.report.create_report(
        project_data,
        collected_imports,
        collected_libs,
        is_strict,
    )

    res = all([not details["brick_diff"], not details["libs_diff"]])

    if not is_quiet:
        check.report.print_missing_deps(details["brick_diff"], name)
        check.report.print_missing_deps(details["libs_diff"], name)

        if is_verbose:
            check.report.print_brick_imports(details["brick_imports"])
            check.report.print_brick_imports(details["third_party_imports"])

    return res



from typing import List

from polylith.reporting import theme
from rich import box
from rich.console import Console
from rich.padding import Padding
from rich.table import Table


def brick_status(brick, bricks, command: str) -> str:
    emoji = theme.check_emoji if command == "info" else ":gear:"

    status = emoji if brick in bricks else "-"

    return f"[data]{status}[/]"


def is_project(project: dict) -> bool:
    return project["type"] == "project"


def printable_name(project: dict, short: bool) -> str:
    if is_project(project):
        template = "[proj]{name}[/]"
        name = project["name"]
    else:
        template = "[data]{name}[/]"
        name = "development"

    if short:
        return template.format(name="\n".join(name))

    return template.format(name=name)


def build_bricks_in_projects_table(
    projects_data: List[dict],
    bases: List[str],
    components: List[str],
    options: dict,
) -> Table:
    short = options.get("short", False)
    command = options.get("command", "info")

    table = Table(box=box.SIMPLE_HEAD)
    table.add_column("[data]brick[/]")

    proj_cols = [printable_name(project, short) for project in projects_data]

    for col in proj_cols:
        table.add_column(col, justify="center")

    for brick in sorted(components):
        statuses = [
            brick_status(brick, p.get("components"), command) for p in projects_data
        ]
        cols = [f"[comp]{brick}[/]"] + statuses

        table.add_row(*cols)

    for brick in sorted(bases):
        statuses = [brick_status(brick, p.get("bases"), command) for p in projects_data]
        cols = [f"[base]{brick}[/]"] + statuses

        table.add_row(*cols)

    return table


def print_table(table: Table) -> None:
    console = Console(theme=theme.poly_theme)

    console.print(table, overflow="ellipsis")


def print_compressed_view_for_bricks_in_projects(
    projects_data: List[dict], bases: List[str], components: List[str]
) -> None:
    options = {"short": True}
    table = build_bricks_in_projects_table(projects_data, bases, components, options)

    print_table(table)


def print_bricks_in_projects(
    projects_data: List[dict], bases: List[str], components: List[str]
) -> None:
    options = {"short": False}
    table = build_bricks_in_projects_table(projects_data, bases, components, options)

    print_table(table)


def print_workspace_summary(
    projects_data: List[dict], bases: List[str], components: List[str]
) -> None:
    console = Console(theme=theme.poly_theme)

    console.print(Padding("[data]Workspace summary[/]", (1, 0, 1, 0)))

    number_of_projects = len([p for p in projects_data if is_project(p)])
    number_of_components = len(components)
    number_of_bases = len(bases)
    number_of_dev = len([p for p in projects_data if not is_project(p)])

    console.print(f"[proj]projects[/]: [data]{number_of_projects}[/]")
    console.print(f"[comp]components[/]: [data]{number_of_components}[/]")
    console.print(f"[base]bases[/]: [data]{number_of_bases}[/]")
    console.print(f"[data]development[/]: [data]{number_of_dev}[/]")



import ast
from functools import lru_cache
from pathlib import Path
from typing import List, Set


def parse_import(node: ast.Import) -> List[str]:
    return [name.name for name in node.names]


def extract_import_from(node: ast.ImportFrom) -> List:
    return (
        [f"{node.module}.{alias.name}" for alias in node.names]
        if node.names
        else [node.module]
    )


def parse_import_from(node: ast.ImportFrom) -> List[str]:
    return extract_import_from(node) if node.module and node.level == 0 else []


def parse_imports(node: ast.AST) -> List[str]:
    if isinstance(node, ast.Import):
        return parse_import(node)

    if isinstance(node, ast.ImportFrom):
        return parse_import_from(node)

    return []


@lru_cache
def parse_module(path: Path) -> ast.AST:
    with open(path.as_posix(), "r", encoding="utf-8", errors="ignore") as f:
        tree = ast.parse(f.read(), path.name)

    return tree


def extract_imports(path: Path) -> List[str]:
    tree = parse_module(path)

    return [i for node in ast.walk(tree) for i in parse_imports(node) if i is not None]


def list_imports(path: Path) -> Set[str]:
    py_modules = path.rglob("*.py")

    extracted = (extract_imports(m) for m in py_modules)
    flattened = (i for imports in extracted for i in imports)

    return set(flattened)


def fetch_all_imports(paths: Set[Path]) -> dict:
    rows = [{p.name: list_imports(p)} for p in paths]

    return {k: v for row in rows for k, v in row.items()}


def extract_top_ns_from_imports(imports: Set[str]) -> Set:
    return {imp.split(".")[0] for imp in imports}


def extract_top_ns(import_data: dict) -> dict:
    return {k: extract_top_ns_from_imports(v) for k, v in import_data.items()}



poetry_pyproject = """\
[tool.poetry]
name = "{name}"
version = "0.1.0"
description = "{description}"
authors = {authors}
license = ""

packages = []

[tool.poetry.dependencies]
python = "{python_version}"

[build-system]
requires = ["poetry-core>=1.0.0"]
build-backend = "poetry.core.masonry.api"
"""

hatch_pyproject = """\
[build-system]
requires = ["hatchling", "hatch-polylith-bricks"]
build-backend = "hatchling.build"

[project]
name = "{name}"
version = "0.1.0"
description = "{description}"
authors = {authors}

requires-python = "{python_version}"

dependencies = []

[tool.hatch.build.hooks.polylith-bricks]

[tool.polylith.bricks]
"""

pdm_pyproject = """\
[build-system]
requires = ["pdm-backend", "pdm-polylith-bricks"]
build-backend = "pdm.backend"

[project]
name = "{name}"
version = "0.1.0"
description = "{description}"
authors = {authors}

requires-python = "{python_version}"

dependencies = []

[tool.polylith.bricks]
"""



from polylith.bricks.base import create_base, get_bases_data
from polylith.bricks.component import create_component, get_components_data

__all__ = [
    "create_base",
    "create_component",
    "get_bases_data",
    "get_components_data",
]



from pathlib import Path

from polylith.files import create_file

keep_file_name = ".keep"


def create_dir(path: Path, dir_name: str, keep=False) -> Path:
    d = path / dir_name
    d.mkdir(parents=True)

    if keep:
        create_file(d, keep_file_name)

    return d



from pathlib import Path

from polylith import configuration, info, repo


def run(short: bool):
    root = repo.get_workspace_root(Path.cwd())

    ns = configuration.get_namespace_from_config(root)
    bases = info.get_bases(root, ns)
    components = info.get_components(root, ns)
    projects_data = info.get_bricks_in_projects(root, components, bases, ns)

    info.print_workspace_summary(projects_data, bases, components)

    if not components and not bases:
        return

    if short:
        info.print_compressed_view_for_bricks_in_projects(
            projects_data, bases, components
        )
    else:
        info.print_bricks_in_projects(projects_data, bases, components)



from pathlib import Path
from typing import List

from polylith.bricks import component
from polylith.bricks.brick import create_brick
from polylith.repo import bases_dir
from polylith.test import create_test


def create_base(path: Path, options: dict) -> None:
    extra = {"brick": bases_dir}
    base_options = {**options, **extra}

    create_brick(path, base_options)
    create_test(path, base_options)


def get_bases_data(path: Path, ns: str) -> List[dict]:
    return component.get_components_data(path, ns, bases_dir)



from cleo.helpers import option
from poetry.console.commands.command import Command
from polylith.bricks import component
from polylith.commands.create import create


class CreateComponentCommand(Command):
    name = "poly create component"
    description = "Creates a <comment>Polylith</> component."

    options = [
        option("name", None, "Name of the component.", flag=False),
        option(
            "description",
            None,
            "Description of the component.",
            flag=False,
            value_required=False,
        ),
    ]

    def handle(self) -> int:
        name = self.option("name")
        description = self.option("description")

        create(name, description, component.create_component)

        return 0



from polylith.poetry.commands.check import CheckCommand
from polylith.poetry.commands.create_base import CreateBaseCommand
from polylith.poetry.commands.create_component import CreateComponentCommand
from polylith.poetry.commands.create_project import CreateProjectCommand
from polylith.poetry.commands.create_workspace import CreateWorkspaceCommand
from polylith.poetry.commands.deps import DepsCommand
from polylith.poetry.commands.diff import DiffCommand
from polylith.poetry.commands.info import InfoCommand
from polylith.poetry.commands.libs import LibsCommand
from polylith.poetry.commands.sync import SyncCommand

__all__ = [
    "CheckCommand",
    "CreateBaseCommand",
    "CreateComponentCommand",
    "CreateProjectCommand",
    "CreateWorkspaceCommand",
    "DepsCommand",
    "DiffCommand",
    "InfoCommand",
    "LibsCommand",
    "SyncCommand",
]



from polylith.project import templates
from polylith.project.create import create_project
from polylith.project.get import get_packages_for_projects, get_project_name, get_toml
from polylith.project.parser import parse_package_paths

__all__ = [
    "create_project",
    "get_packages_for_projects",
    "get_project_name",
    "get_toml",
    "parse_package_paths",
    "templates",
]



from functools import reduce
from pathlib import Path
from typing import List, Set

from polylith import distributions
from polylith.libs import report


def missing_libs(project_data: dict, imports: dict, options: dict) -> bool:
    is_strict = options["strict"]
    library_alias = options["alias"]

    name = project_data["name"]
    deps = project_data["deps"]

    brick_imports = imports[name]

    libs = distributions.known_aliases_and_sub_dependencies(
        deps, library_alias, options
    )

    return report.print_missing_installed_libs(
        brick_imports,
        libs,
        name,
        is_strict,
    )


def flatten_imports(acc: dict, item: dict) -> dict:
    bases = item.get("bases", {})
    components = item.get("components", {})

    return {
        "bases": {**acc.get("bases", {}), **bases},
        "components": {**acc.get("components", {}), **components},
    }


def run_library_versions(
    projects_data: List[dict], all_projects_data: List[dict], options: dict
) -> None:
    development_data = next(p for p in all_projects_data if p["type"] == "development")
    filtered_projects_data = [p for p in projects_data if p["type"] != "development"]

    report.print_libs_in_projects(development_data, filtered_projects_data, options)


def run(
    root: Path,
    ns: str,
    projects_data: List[dict],
    options: dict,
) -> Set[bool]:
    imports = {
        p["name"]: report.get_third_party_imports(root, ns, p) for p in projects_data
    }

    flattened: dict = reduce(flatten_imports, imports.values(), {})

    report.print_libs_summary()
    report.print_libs_in_bricks(flattened)

    return {missing_libs(p, imports, options) for p in projects_data}



from pathlib import Path
from typing import Set

from polylith import imports, libs, workspace
from polylith.check import collect, grouping
from polylith.reporting import theme
from rich.console import Console


def print_brick_imports(brick_imports: dict) -> None:
    console = Console(theme=theme.poly_theme)

    bases = brick_imports["bases"]
    components = brick_imports["components"]

    bricks = {**bases, **components}

    for key, values in bricks.items():
        imports_in_brick = values.difference({key})

        if not imports_in_brick:
            continue

        joined = ", ".join(imports_in_brick)
        message = f":information: [data]{key}[/] is importing [data]{joined}[/]"
        console.print(message)


def print_missing_deps(diff: Set[str], project_name: str) -> None:
    if not diff:
        return

    console = Console(theme=theme.poly_theme)

    missing = ", ".join(sorted(diff))

    console.print(f":thinking_face: Cannot locate {missing} in {project_name}")


def fetch_brick_imports(root: Path, ns: str, all_imports: dict) -> dict:
    extracted = grouping.extract_brick_imports(all_imports, ns)

    return collect.with_unknown_components(root, ns, extracted)


def collect_all_imports(root: Path, ns: str, project_data: dict) -> dict:
    bases = {b for b in project_data.get("bases", [])}
    components = {c for c in project_data.get("components", [])}

    bases_paths = workspace.paths.collect_bases_paths(root, ns, bases)
    components_paths = workspace.paths.collect_components_paths(root, ns, components)

    all_imports_in_bases = imports.fetch_all_imports(bases_paths)
    all_imports_in_components = imports.fetch_all_imports(components_paths)

    brick_imports = {
        "bases": fetch_brick_imports(root, ns, all_imports_in_bases),
        "components": fetch_brick_imports(root, ns, all_imports_in_components),
    }

    third_party_imports = {
        "bases": libs.extract_third_party_imports(all_imports_in_bases, ns),
        "components": libs.extract_third_party_imports(all_imports_in_components, ns),
    }

    return {"brick_imports": brick_imports, "third_party_imports": third_party_imports}


def create_report(
    project_data: dict,
    collected_imports: dict,
    third_party_libs: Set,
    is_strict: bool = False,
) -> dict:
    bases = {b for b in project_data.get("bases", [])}
    components = {c for c in project_data.get("components", [])}

    brick_imports = collected_imports["brick_imports"]
    third_party_imports = collected_imports["third_party_imports"]

    brick_diff = collect.imports_diff(brick_imports, bases, components)
    libs_diff = libs.report.calculate_diff(
        third_party_imports, third_party_libs, is_strict
    )

    return {
        "brick_imports": brick_imports,
        "third_party_imports": third_party_imports,
        "brick_diff": brick_diff,
        "libs_diff": libs_diff,
    }



from polylith.configuration.core import (
    get_brick_structure_from_config,
    get_namespace_from_config,
    get_resources_structure_from_config,
    get_tag_pattern_from_config,
    get_tag_sort_options_from_config,
    get_tests_structure_from_config,
    get_theme_from_config,
    is_readme_generation_enabled,
    is_test_generation_enabled,
)

__all__ = [
    "get_brick_structure_from_config",
    "get_namespace_from_config",
    "get_resources_structure_from_config",
    "get_tag_pattern_from_config",
    "get_tag_sort_options_from_config",
    "get_tests_structure_from_config",
    "get_theme_from_config",
    "is_readme_generation_enabled",
    "is_test_generation_enabled",
]



from pathlib import Path
from typing import Set

from polylith import check, workspace


def get_brick_imports(
    root: Path, ns: str, bases: Set[str], components: Set[str]
) -> dict:
    bases_paths = workspace.paths.collect_bases_paths(root, ns, bases)
    comp_paths = workspace.paths.collect_components_paths(root, ns, components)

    brick_imports_in_bases = check.collect.extract_bricks(bases_paths, ns)
    brick_imports_in_components = check.collect.extract_bricks(comp_paths, ns)

    return {
        "bases": check.collect.with_unknown_components(
            root, ns, brick_imports_in_bases
        ),
        "components": check.collect.with_unknown_components(
            root, ns, brick_imports_in_components
        ),
    }



from pathlib import Path

from cleo.helpers import option
from poetry.console.commands.command import Command
from polylith import project
from polylith.commands.create import create

command_name = "poly create project"


def create_project(root: Path, options: dict):
    package = options["package"]
    desc = options["description"] or ""

    project.create_project(root, project.templates.poetry_pyproject, package, desc)


class CreateProjectCommand(Command):
    name = command_name
    description = "Creates a <comment>Polylith</> project."

    options = [
        option("name", None, "Name of the project.", flag=False),
        option(
            "description",
            None,
            "Description of the project.",
            flag=False,
            value_required=False,
        ),
    ]

    def handle(self) -> int:
        name = self.option("name")
        description = self.option("description")

        create(name, description, create_project)

        return 0



from pathlib import Path

from polylith.files import create_file

template_docstring = """\
\"\"\"
{description}
\"\"\"

"""

template_content = """\
from {namespace}.{package} import {modulename}

__all__ = ["{modulename}"]
"""


def to_namespaced_path(package: str) -> str:
    parts = package.split("/")

    return ".".join(parts)


def create_interface(path: Path, options: dict) -> None:
    interface = create_file(path, "__init__.py")

    namespace = options["namespace"]
    package = options["package"]
    description = options["description"]
    modulename = options["modulename"]

    package_path = to_namespaced_path(package)

    docstring = (
        template_docstring.format(description=description) if description else ""
    )

    content = template_content.format(
        namespace=namespace, package=package_path, modulename=modulename
    )

    interface.write_text(docstring + content)



from poetry.console.application import Application
from poetry.plugins.application_plugin import ApplicationPlugin
from polylith.poetry.commands import (
    CheckCommand,
    CreateBaseCommand,
    CreateComponentCommand,
    CreateProjectCommand,
    CreateWorkspaceCommand,
    DepsCommand,
    DiffCommand,
    InfoCommand,
    LibsCommand,
    SyncCommand,
)

commands = [
    CheckCommand,
    CreateBaseCommand,
    CreateComponentCommand,
    CreateProjectCommand,
    CreateWorkspaceCommand,
    DepsCommand,
    DiffCommand,
    InfoCommand,
    LibsCommand,
    SyncCommand,
]


def register_command(application: Application, command) -> None:
    application.command_loader.register_factory(command.name, command)


def register_commands(application: Application) -> None:
    for command in commands:
        register_command(application, command)


class PolylithPlugin(ApplicationPlugin):
    def activate(self, application: Application):
        register_commands(application)



from typing import Set, Union


def only_brick_imports(imports: Set[str], top_ns: str) -> Set[str]:
    return {i for i in imports if i.startswith(top_ns)}


def only_bricks(import_data: dict, top_ns: str) -> dict:
    return {k: only_brick_imports(v, top_ns) for k, v in import_data.items()}


def brick_import_to_name(brick_import: str) -> Union[str, None]:
    parts = brick_import.split(".")

    return parts[1] if len(parts) > 1 else None


def only_brick_name(brick_imports: Set[str]) -> Set:
    res = {brick_import_to_name(i) for i in brick_imports}

    return {i for i in res if i}


def only_brick_names(import_data: dict) -> dict:
    return {k: only_brick_name(v) for k, v in import_data.items() if v}


def exclude_empty(import_data: dict) -> dict:
    return {k: v for k, v in import_data.items() if v}


def extract_brick_imports(all_imports: dict, top_ns) -> dict:
    with_only_bricks = only_bricks(all_imports, top_ns)
    with_only_brick_names = only_brick_names(with_only_bricks)

    return exclude_empty(with_only_brick_names)



from pathlib import Path
from typing import List, Set, Union

from polylith import bricks, deps, info


def get_imports(root: Path, ns: str, bases: Set[str], components: Set[str]) -> dict:
    brick_imports = deps.get_brick_imports(root, ns, bases, components)

    return {**brick_imports["bases"], **brick_imports["components"]}


def pick_name(data: List[dict]) -> Set[str]:
    return {b["name"] for b in data}


def get_bases(root: Path, ns: str, project_data: dict) -> Set[str]:
    if project_data:
        return set(project_data.get("bases", []))

    return pick_name(bricks.get_bases_data(root, ns))


def get_components(root: Path, ns: str, project_data: dict) -> Set[str]:
    if project_data:
        return set(project_data.get("components", []))

    return pick_name(bricks.get_components_data(root, ns))


def run(root: Path, ns: str, directory: Union[str, None], brick: Union[str, None]):
    projects_data = info.get_projects_data(root, ns) if directory else []
    project = next((p for p in projects_data if directory in p["path"].as_posix()), {})

    bases = get_bases(root, ns, project)
    components = get_components(root, ns, project)

    imports = get_imports(root, ns, bases, components)

    if brick and imports.get(brick):
        deps.print_brick_deps(brick, bases, components, imports)
        return

    deps.print_deps(bases, components, imports)



from polylith.info.collect import (
    get_bases,
    get_bricks_in_projects,
    get_components,
    get_projects_data,
)
from polylith.info.report import (
    is_project,
    print_bricks_in_projects,
    print_compressed_view_for_bricks_in_projects,
    print_workspace_summary,
)

__all__ = [
    "get_bases",
    "get_bricks_in_projects",
    "get_components",
    "get_projects_data",
    "is_project",
    "print_bricks_in_projects",
    "print_compressed_view_for_bricks_in_projects",
    "print_workspace_summary",
]



from pathlib import Path
from typing import List

from polylith.bricks import base, component
from polylith.project import get_packages_for_projects, parse_package_paths


def get_matching_bricks(
    paths: List[Path], bricks: List[str], namespace: str
) -> List[str]:
    paths_in_namespace = (p.name for p in paths if p.parent.name == namespace)

    res = set(bricks).intersection(paths_in_namespace)

    return sorted(list(res))


def get_project_bricks(project_packages: List[dict], components, bases, namespace: str):
    paths = parse_package_paths(project_packages)

    components_in_project = get_matching_bricks(paths, components, namespace)
    bases_in_project = get_matching_bricks(paths, bases, namespace)

    return {"components": components_in_project, "bases": bases_in_project}


def get_components(root: Path, namespace: str) -> List[str]:
    return [c["name"] for c in component.get_components_data(root, namespace)]


def get_bases(root: Path, namespace: str) -> List[str]:
    return [b["name"] for b in base.get_bases_data(root, namespace)]


def get_bricks_in_projects(
    root: Path, components: List[str], bases: List[str], namespace: str
) -> List[dict]:
    packages_for_projects = get_packages_for_projects(root)

    res = [
        {
            **{k: v for k, v in p.items() if k not in {"packages"}},
            **get_project_bricks(p["packages"], components, bases, namespace),
        }
        for p in packages_for_projects
    ]

    return res


def get_projects_data(root: Path, ns: str) -> List[dict]:
    bases = get_bases(root, ns)
    components = get_components(root, ns)

    return get_bricks_in_projects(root, components, bases, ns)



from pathlib import Path

from polylith.toml import load_toml

patterns = {
    "pdm.lock": "toml",
    "poetry.lock": "toml",
    "requirements.lock": "text",
    "requirements.txt": "text",
}


def find_lock_files(path: Path) -> dict:
    return {k: v for k, v in patterns.items() if Path(path / k).exists()}


def pick_lock_file(project_data: dict) -> dict:
    data = find_lock_files(project_data["path"])
    first = next(iter(data.items()), None)

    if not first:
        return {}

    filename, filetype = first

    return {"filename": filename, "filetype": filetype}


def extract_lib_names_from_toml(path: Path) -> dict:
    data = load_toml(path)

    return {p["name"]: p["version"] for p in data.get("package", [])}


def parse_name(row: str) -> str:
    parts = str.split(row, "==")

    return parts[0]


def parse_version(row: str) -> str:
    parts = str.split(row, "==")[1]
    res = str.split(parts, " ")

    return res[0]


def extract_lib_names_from_txt(path: Path) -> dict:
    with open(path, "r") as f:
        data = f.readlines()

    rows = (str.strip(line) for line in data)
    filtered = (row for row in rows if row and not row.startswith(("#", "-")))

    return {parse_name(row): parse_version(row) for row in filtered}


def extract_libs(project_data: dict, filename: str, filetype: str) -> dict:
    dir_path = project_data["path"]

    path = Path(dir_path / filename)

    if not path.exists():
        return {}

    try:
        if filetype == "toml":
            return extract_lib_names_from_toml(path)

        return extract_lib_names_from_txt(path)
    except (IndexError, KeyError, ValueError) as e:
        raise ValueError(f"Failed reading {filename}: {repr(e)}") from e



from polylith.repo.get import get_authors, get_python_version
from polylith.repo.repo import (
    bases_dir,
    components_dir,
    default_toml,
    development_dir,
    get_workspace_root,
    is_hatch,
    is_pdm,
    is_pep_621_ready,
    is_poetry,
    projects_dir,
    readme_file,
    workspace_file,
)

__all__ = [
    "get_authors",
    "get_python_version",
    "bases_dir",
    "components_dir",
    "default_toml",
    "development_dir",
    "get_workspace_root",
    "is_hatch",
    "is_pdm",
    "is_pep_621_ready",
    "is_poetry",
    "projects_dir",
    "readme_file",
    "workspace_file",
]



from functools import reduce
from itertools import zip_longest
from typing import List, Set, Tuple

from polylith.reporting import theme
from rich import box
from rich.console import Console
from rich.table import Table


def to_col(brick: str, tag: str) -> str:
    name = "\n".join(brick)

    return f"[{tag}]{name}[/]"


def brick_status(bricks: Set[str], brick_name: str, imported: str) -> str:
    status = theme.check_emoji if imported in bricks and imported != brick_name else "-"

    return f"[data]{status}[/]"


def to_row(name: str, tag: str, brick_imports: dict, imported: List[str]) -> List[str]:
    bricks = brick_imports.get(name) or set()
    statuses = [brick_status(bricks, name, i) for i in imported]

    return [f"[{tag}]{name}[/]"] + statuses


def flatten_import(acc: Set[str], kv: Tuple[str, Set[str]]) -> set:
    key = kv[0]
    values = kv[1]

    return set().union(acc, values.difference({key}))


def flatten_imports(brick_imports: dict) -> Set[str]:
    """Flatten the dict into a set of imports, with the actual brick filtered away when existing as an import"""
    return reduce(flatten_import, brick_imports.items(), set())


def create_columns(
    imported_bases: List[str], imported_components: List[str]
) -> List[str]:
    base_cols = [to_col(brick, "base") for brick in imported_bases]
    comp_cols = [to_col(brick, "comp") for brick in imported_components]

    return comp_cols + base_cols


def create_rows(
    bases: Set[str], components: Set[str], import_data: dict, imported: List[str]
) -> List[List[str]]:
    base_rows = [to_row(b, "base", import_data, imported) for b in sorted(bases)]
    comp_rows = [to_row(c, "comp", import_data, imported) for c in sorted(components)]

    return comp_rows + base_rows


def print_deps(bases: Set[str], components: Set[str], import_data: dict):
    flattened = flatten_imports(import_data)

    imported_bases = sorted({b for b in flattened if b in bases})
    imported_components = sorted({c for c in flattened if c in components})
    imported_bricks = imported_components + imported_bases

    table = Table(box=box.SIMPLE_HEAD)
    table.add_column("[data]brick[/]")

    cols = create_columns(imported_bases, imported_components)
    rows = create_rows(bases, components, import_data, imported_bricks)

    for col in cols:
        table.add_column(col, justify="center")

    for row in rows:
        table.add_row(*row)

    console = Console(theme=theme.poly_theme)
    console.print(table, overflow="ellipsis")


def without(key: str, bricks: Set[str]) -> Set[str]:
    return {b for b in bricks if b != key}


def sorted_usings(usings: Set[str], bases: Set[str], components: Set[str]) -> List[str]:
    usings_bases = sorted({b for b in usings if b in bases})
    usings_components = sorted({c for c in usings if c in components})

    return usings_components + usings_bases


def sorted_used_by(
    brick: str, bases: Set[str], components: Set[str], import_data: dict
) -> List[str]:
    brick_used_by = without(brick, {k for k, v in import_data.items() if brick in v})

    return sorted_usings(brick_used_by, bases, components)


def sorted_uses(
    brick: str, bases: Set[str], components: Set[str], import_data: dict
) -> List[str]:
    brick_uses = without(brick, {b for b in import_data[brick]})

    return sorted_usings(brick_uses, bases, components)


def calculate_tag(brick: str, bases: Set[str]) -> str:
    return "base" if brick in bases else "comp"


def print_brick_deps(
    brick: str, bases: Set[str], components: Set[str], import_data: dict
):
    brick_used_by = sorted_used_by(brick, bases, components, import_data)
    brick_uses = sorted_uses(brick, bases, components, import_data)

    tag = calculate_tag(brick, bases)

    console = Console(theme=theme.poly_theme)

    table = Table(box=box.SIMPLE_HEAD)
    table.add_column("[data]used by[/]")
    table.add_column(":backhand_index_pointing_left:")
    table.add_column(f"[{tag}]{brick}[/]")
    table.add_column(":backhand_index_pointing_right:")
    table.add_column("[data]uses[/]")

    for item in zip_longest(brick_used_by, brick_uses):
        used_by, uses = item

        used_by_tag = calculate_tag(used_by, bases) if used_by else ""
        uses_tag = calculate_tag(uses, bases) if uses else ""

        left = f"[{used_by_tag}]{used_by}[/]" if used_by else ""
        right = f"[{uses_tag}]{uses}[/]" if uses else ""

        row = [left, "", "", "", right]

        table.add_row(*row)

    console.print(table, overflow="ellipsis")



from typing import List, Set

from polylith import info
from polylith.reporting import theme
from rich.console import Console
from rich.padding import Padding


def print_diff_details(
    projects_data: List[dict], bases: List[str], components: List[str]
) -> None:
    if not bases and not components:
        return

    console = Console(theme=theme.poly_theme)

    options = {"command": "diff"}
    table = info.report.build_bricks_in_projects_table(
        projects_data, bases, components, options
    )

    console.print(table, overflow="ellipsis")


def print_detected_changes(changes: List[str], markup: str, short: bool) -> None:
    if not changes:
        return

    console = Console(theme=theme.poly_theme)

    if short:
        console.out(",".join(changes))
        return

    for brick in changes:
        console.print(f"[data]:gear: Changes found in [/][{markup}]{brick}[/]")


def print_detected_dependent(bases: List[str], components: List[str]) -> None:
    console = Console(theme=theme.poly_theme)

    printable_bases = [f"[base]{b}[/]" for b in bases]
    printable_components = [f"[comp]{c}[/]" for c in components]

    printable_bricks = printable_components + printable_bases
    joined = ", ".join(printable_bricks) or "-"

    console.print(f"[data]:gear: Used by: [/]{joined}")


def print_detected_changes_in_bricks(
    changed_bases: List[str],
    changed_components: List[str],
    dependent_bricks: dict,
    options: dict,
) -> None:
    short = options.get("short", False)
    with_deps = options.get("deps", False)

    dependent_bases = sorted(dependent_bricks.get("bases", set()))
    dependent_components = sorted(dependent_bricks.get("components", set()))

    bricks = changed_components + changed_bases + dependent_components + dependent_bases

    if short:
        print_detected_changes(bricks, "data", short)
        return

    print_detected_changes(changed_components, "comp", short)
    print_detected_changes(changed_bases, "base", short)

    if with_deps:
        print_detected_dependent(dependent_bases, dependent_components)


def print_detected_changes_in_projects(projects: List[str], short: bool) -> None:
    print_detected_changes(projects, "proj", short)


def print_projects_affected_by_changes(projects: Set[str], short: bool) -> None:
    sorted_projects = sorted(list(projects))

    print_detected_changes(sorted_projects, "proj", short)


def print_diff_summary(tag: str, bases: List[str], components: List[str]) -> None:
    console = Console(theme=theme.poly_theme)

    console.print(Padding(f"[data]Diff: based on the {tag} tag[/]", (1, 0, 1, 0)))

    if not bases and not components:
        console.print("[data]No brick changes found.[/]")
        return

    if components:
        console.print(f"[comp]Changed components[/]: [data]{len(components)}[/]")

    if bases:
        console.print(f"[base]Changed bases[/]: [data]{len(bases)}[/]")



from pathlib import Path

from polylith import check, deps, info


def calculate_diff(
    root: Path,
    namespace: str,
    project_data: dict,
    workspace_data: dict,
) -> dict:
    bases = set(project_data["bases"])
    components = set(project_data["components"])

    all_bases = workspace_data["bases"]
    all_components = workspace_data["components"]

    brick_imports = deps.get_brick_imports(root, namespace, bases, components)
    is_project = info.is_project(project_data)

    if is_project:
        brick_diff = check.collect.imports_diff(brick_imports, bases, components)
    else:
        all_bricks = set().union(all_bases, all_components)
        brick_diff = check.collect.diff(all_bricks, bases, components)

    bases_diff = {b for b in brick_diff if b in all_bases}
    components_diff = {b for b in brick_diff if b in all_components}

    return {
        "name": project_data["name"],
        "path": project_data["path"],
        "is_project": is_project,
        "bases": bases_diff,
        "components": components_diff,
        "brick_imports": brick_imports,
    }



from pathlib import Path

from poetry.console.commands.command import Command
from polylith import commands, configuration, info, repo
from polylith.poetry.internals import filter_projects_data


class SyncCommand(Command):
    name = "poly sync"
    description = "Update <comment>pyproject.toml</comment> with missing bricks."

    def handle(self) -> int:
        directory = self.option("directory")
        root = repo.get_workspace_root(Path.cwd())
        ns = configuration.get_namespace_from_config(root)

        all_projects_data = info.get_projects_data(root, ns)
        projects_data = filter_projects_data(self.poetry, directory, all_projects_data)

        options = {"verbose": self.option("verbose"), "quiet": self.option("quiet")}

        for data in projects_data:
            commands.sync.run(root, ns, data, options)

        return 0



from pathlib import Path
from typing import List, Set, Union

from polylith import configuration, deps, diff, info, repo


def get_imports(root: Path, ns: str, bases: Set[str], components: Set[str]) -> dict:
    brick_imports = deps.get_brick_imports(root, ns, bases, components)

    return {**brick_imports["bases"], **brick_imports["components"]}


def get_projects_data(root: Path, ns: str) -> List[dict]:
    data = info.get_projects_data(root, ns)

    return [p for p in data if info.is_project(p)]


def flatten_bricks(projects_data: List[dict], brick_type: str) -> Set[str]:
    matrix = [p[brick_type] for p in projects_data]

    flattened: List[str] = sum(matrix, [])

    return set(flattened)


def flatten_dependent_bricks(
    changed_bricks: Set[str], bases: Set[str], components: Set[str], import_data: dict
) -> Set[str]:
    matrix = [
        deps.report.sorted_used_by(brick, bases, components, import_data)
        for brick in changed_bricks
    ]

    flattened: List[str] = sum(matrix, [])
    filtered = set(flattened).difference(changed_bricks)

    return filtered


def calculate_dependent_bricks(
    root: Path, ns: str, projects_data: List[dict], changed_bricks: Set[str]
) -> dict:
    bases = flatten_bricks(projects_data, "bases")
    components = flatten_bricks(projects_data, "components")
    import_data = get_imports(root, ns, bases, components)

    dependent_bricks = flatten_dependent_bricks(
        changed_bricks, bases, components, import_data
    )

    dependent_bases = {b for b in dependent_bricks if b in bases}
    dependent_components = {c for c in dependent_bricks if c in components}

    return {"bases": dependent_bases, "components": dependent_components}


def print_views(root: Path, tag: str, options: dict) -> None:
    short = options.get("short", False)
    only_bricks = options.get("bricks", False)
    with_deps = options.get("deps", False)

    ns = configuration.get_namespace_from_config(root)
    files = diff.collect.get_files(tag)

    changed_bases = diff.collect.get_changed_bases(root, files, ns)
    changed_components = diff.collect.get_changed_components(root, files, ns)
    changed_bricks = set(changed_bases + changed_components)
    changed_projects = diff.collect.get_changed_projects(files)

    projects_data = get_projects_data(root, ns)

    if not short and not only_bricks:
        diff.report.print_diff_summary(tag, changed_bases, changed_components)
        diff.report.print_detected_changes_in_projects(changed_projects, short)
        diff.report.print_diff_details(projects_data, changed_bases, changed_components)

        return

    if short and not only_bricks:
        affected_projects = diff.collect.get_projects_affected_by_changes(
            projects_data, changed_projects, changed_bases, changed_components
        )
        diff.report.print_projects_affected_by_changes(affected_projects, short)

        return

    dependent_bricks = (
        calculate_dependent_bricks(root, ns, projects_data, changed_bricks)
        if with_deps
        else {}
    )

    diff.report.print_detected_changes_in_bricks(
        changed_bases, changed_components, dependent_bricks, options
    )


def run(tag_name: Union[str, None], options: dict):
    root = repo.get_workspace_root(Path.cwd())

    tag = diff.collect.get_latest_tag(root, tag_name)

    if not tag:
        print("No tags found in repository.")
        return

    print_views(root, tag, options)



from pathlib import Path

import tomlkit
from polylith import readme, repo
from polylith.development import create_development
from polylith.dirs import create_dir

template = """\
[tool.polylith]
namespace = "{namespace}"
git_tag_pattern = "stable-*"

[tool.polylith.structure]
theme = "{theme}"

[tool.polylith.tag.patterns]
stable = "stable-*"
release = "v[0-9]*"

[tool.polylith.resources]
brick_docs_enabled = false

[tool.polylith.test]
enabled = true
"""


def create_workspace_config(path: Path, namespace: str, theme: str) -> None:
    formatted = template.format(namespace=namespace, theme=theme)
    content: dict = tomlkit.loads(formatted)

    fullpath = path / repo.workspace_file

    with fullpath.open("w", encoding="utf-8") as f:
        f.write(tomlkit.dumps(content))


def create_workspace(path: Path, namespace: str, theme: str) -> None:
    create_dir(path, repo.bases_dir, keep=True)
    create_dir(path, repo.components_dir, keep=True)
    create_dir(path, repo.projects_dir, keep=True)

    create_development(path, keep=True)

    create_workspace_config(path, namespace, theme)

    readme.create_workspace_readme(path, namespace)



from functools import reduce
from typing import Dict, List, Set


def _to_key_with_values(acc: Dict, alias: str) -> Dict:
    k, v = str.split(alias, "=")

    values = [str.strip(val) for val in str.split(v, ",")]

    return {**acc, **{k: values}}


def parse(aliases: List[str]) -> Dict[str, List[str]]:
    """Parse a list of aliases defined as key=value(s) into a dictionary"""
    return reduce(_to_key_with_values, aliases, {})


def pick(aliases: Dict[str, List[str]], keys: Set) -> Set:
    normalized_keys = {str.lower(k) for k in keys}

    matrix = [v for k, v in aliases.items() if str.lower(k) in normalized_keys]

    flattened: List = sum(matrix, [])

    return set(flattened)



from polylith.distributions.collect import known_aliases_and_sub_dependencies
from polylith.distributions.core import (
    distributions_packages,
    distributions_sub_packages,
    get_distributions,
)

__all__ = [
    "distributions_packages",
    "distributions_sub_packages",
    "get_distributions",
    "known_aliases_and_sub_dependencies",
]



from polylith.libs import report
from polylith.libs.grouping import extract_third_party_imports, get_third_party_imports
from polylith.libs.lock_files import extract_libs, pick_lock_file

__all__ = [
    "report",
    "extract_third_party_imports",
    "get_third_party_imports",
    "extract_libs",
    "pick_lock_file",
]



from polylith import check
from polylith.reporting import theme
from rich.console import Console


def print_brick_imports(diff: dict) -> None:
    brick_imports = diff["brick_imports"]

    check.report.print_brick_imports(brick_imports)


def print_summary(diff: dict) -> None:
    console = Console(theme=theme.poly_theme)

    is_project = diff["is_project"]
    name = diff["name"] if is_project else "development"
    bases = diff["bases"]
    components = diff["components"]

    anything_to_sync = bases or components

    emoji = ":point_right:" if anything_to_sync else theme.check_emoji
    printable_name = f"[proj]{name}[/]" if is_project else f"[data]{name}[/]"

    console.print(f"{emoji} {printable_name}")

    for b in bases:
        console.print(f"adding [base]{b}[/] base to [proj]{name}[/]")

    for c in components:
        console.print(f"adding [comp]{c}[/] component to [proj]{name}[/]")

    if anything_to_sync:
        console.print("")



from pathlib import Path
from typing import Set

from polylith import check, imports, workspace


def extract_bricks(paths: Set[Path], ns: str) -> dict:
    all_imports = imports.fetch_all_imports(paths)

    return check.grouping.extract_brick_imports(all_imports, ns)


def with_unknown_components(root: Path, ns: str, brick_imports: dict) -> dict:
    keys = set(brick_imports.keys())
    values = set().union(*brick_imports.values())

    unknowns = values.difference(keys)

    if not unknowns:
        return brick_imports

    paths = workspace.paths.collect_components_paths(root, ns, unknowns)

    extracted = extract_bricks(paths, ns)

    if not extracted:
        return brick_imports

    collected = {**brick_imports, **extracted}

    return with_unknown_components(root, ns, collected)


def diff(known_bricks: Set[str], bases: Set[str], components: Set[str]) -> Set[str]:
    bricks = set().union(bases, components)

    return known_bricks.difference(bricks)


def imports_diff(
    brick_imports: dict, bases: Set[str], components: Set[str]
) -> Set[str]:
    flattened_bases = set().union(*brick_imports["bases"].values())
    flattened_components = set().union(*brick_imports["components"].values())

    flattened_imports = set().union(flattened_bases, flattened_components)

    return diff(flattened_imports, bases, components)



from pathlib import Path

from cleo.helpers import option
from poetry.console.commands.command import Command
from polylith.workspace.create import create_workspace


class CreateWorkspaceCommand(Command):
    name = "poly create workspace"
    description = "Creates a <comment>Polylith</> workspace in the current directory."

    options = [
        option(long_name="name", description="Name of the workspace.", flag=False),
        option(
            long_name="theme",
            description="Workspace theme",
            flag=False,
            default="tdd",
        ),
    ]

    def handle(self) -> int:
        path = Path.cwd()
        namespace = self.option("name")
        theme = self.option("theme")

        if not namespace:
            raise ValueError(
                "Please add a workspace name. Poetry poly create workspace --name myname"
            )

        create_workspace(path, namespace, theme)

        return 0



from pathlib import Path
from cleo.helpers import option

from poetry.console.commands.command import Command
from polylith import commands, configuration, repo


class DepsCommand(Command):
    name = "poly deps"
    description = "Visualize the dependencies between <comment>bricks</>."

    options = [
        option(
            long_name="brick",
            description="Shows dependencies for selected brick",
            flag=False,
        ),
    ]

    def handle(self) -> int:
        directory = self.option("directory")
        brick = self.option("brick")

        root = repo.get_workspace_root(Path.cwd())
        ns = configuration.get_namespace_from_config(root)

        dir_path = Path(directory).as_posix() if directory else None

        commands.deps.run(root, ns, dir_path, brick)

        return 0



import importlib.metadata
import re
from functools import lru_cache, reduce
from typing import Dict, List

SUB_DEP_SEPARATORS = r"[\s!=;><\^~]"


def parse_sub_package_name(dependency: str) -> str:
    parts = re.split(SUB_DEP_SEPARATORS, dependency)

    return str(parts[0])


def dist_subpackages(dist) -> dict:
    name = dist.metadata["name"]
    dependencies = importlib.metadata.requires(name) or []

    parsed_package_names = list({parse_sub_package_name(d) for d in dependencies})

    return {name: parsed_package_names} if dependencies else {}


def map_sub_packages(acc, dist) -> dict:
    return {**acc, **dist_subpackages(dist)}


def parsed_namespaces_from_files(dist) -> List[str]:
    name = dist.metadata["name"]
    files = dist.files or []

    normalized_name = str.replace(name, "-", "_")
    to_ignore = {
        name,
        normalized_name,
        str.lower(name),
        str.lower(normalized_name),
        "..",
    }

    filtered = [f for f in files if f.suffix == ".py"]
    top_folders = {f.parts[0] for f in filtered if len(f.parts) > 1}
    namespaces = {t for t in top_folders if t not in to_ignore}

    return list(namespaces)


def parsed_top_level_namespace(namespaces: List[str]) -> List[str]:
    return [str.replace(ns, "/", ".") for ns in namespaces]


def top_level_packages(dist) -> List[str]:
    top_level = dist.read_text("top_level.txt")

    namespaces = str.split(top_level or "")

    return parsed_top_level_namespace(namespaces) or parsed_namespaces_from_files(dist)


def mapped_packages(dist) -> dict:
    packages = top_level_packages(dist)
    name = dist.metadata["name"]

    return {name: packages} if packages else {}


def map_packages(acc, dist) -> dict:
    return {**acc, **mapped_packages(dist)}


def distributions_packages(dists) -> Dict[str, List[str]]:
    """Return a mapping of top-level packages to their distributions."""
    return reduce(map_packages, dists, {})


def distributions_sub_packages(dists) -> Dict[str, List[str]]:
    """Return the dependencies of each distribution."""
    return reduce(map_sub_packages, dists, {})


@lru_cache
def get_distributions() -> list:
    return list(importlib.metadata.distributions())


def get_packages_distributions(project_dependencies: set) -> set:
    """Return the mapped top namespace from an import

    Example:
    A third-party library, such as opentelemetry-instrumentation-fastapi.
    The return value would be the mapped top namespace: opentelemetry

    Note: available for Python >= 3.10
    """

    # added in Python 3.10
    fn = getattr(importlib.metadata, "packages_distributions", None)

    dists = fn() if fn else {}

    common = {k for k, v in dists.items() if project_dependencies.intersection(set(v))}

    return common.difference(project_dependencies)



from functools import lru_cache
from pathlib import Path
from typing import List, Union

from poetry.factory import Factory
from poetry.poetry import Poetry
from poetry.utils.env import EnvManager
from polylith import project


def get_project_poetry(poetry: Poetry, path: Union[Path, None]) -> Poetry:
    return Factory().create_poetry(path) if path else poetry


@lru_cache
def distributions(poetry: Poetry, path: Union[Path, None]) -> list:
    """Get distributions from the current Poetry context.

    When running code within Poetry, the current environment is the one Poetry uses and
    not the environment in the current project or workspace.

    Querying importlib.metadata will fail to find the libraries added to the workspace.

    This function uses the Poetry internals to fetch the distributions,
    that internally queries the metadata based on the current path.
    """
    project_poetry = get_project_poetry(poetry, path)

    env = EnvManager(project_poetry).get()

    return list(env.site_packages.distributions())


def find_third_party_libs(poetry: Poetry, path: Union[Path, None]) -> dict:
    project_poetry = get_project_poetry(poetry, path)

    if not project_poetry.locker.is_locked():
        raise ValueError("poetry.lock not found. Run `poetry lock` to create it.")

    packages = project_poetry.locker.locked_repository().packages

    return {p.name: str(p.version) for p in packages}


def filter_projects_data(
    poetry: Poetry, directory: Union[str, None], projects_data: List[dict]
) -> List[dict]:
    if not directory:
        return projects_data

    project_name = project.get_project_name(poetry.pyproject.data)

    data = next((p for p in projects_data if p["name"] == project_name), None)

    if not data:
        raise ValueError(f"Didn't find project in {directory}")

    return [data]



from cleo.helpers import option
from poetry.console.commands.command import Command
from polylith.bricks import base
from polylith.commands.create import create


class CreateBaseCommand(Command):
    name = "poly create base"
    description = "Creates a <comment>Polylith</> base."

    options = [
        option("name", None, "Name of the base.", flag=False),
        option(
            "description",
            None,
            "Description of the base.",
            flag=False,
            value_required=False,
        ),
    ]

    def handle(self) -> int:
        name = self.option("name")
        description = self.option("description")

        create(name, description, base.create_base)

        return 0



from pathlib import Path
from typing import List


def to_path(package: dict) -> Path:
    include = package["include"]
    from_path = package.get("from")

    return Path(f"{from_path}/{include}") if from_path else Path(include)


def parse_package_paths(packages: List[dict]) -> List[Path]:
    sorted_packages = sorted(packages, key=lambda p: (p.get("from", "."), p["include"]))

    return [to_path(p) for p in sorted_packages]



from pathlib import Path
from typing import List

from polylith import configuration
from polylith.bricks.brick import create_brick
from polylith.repo import components_dir
from polylith.test import create_test


def create_component(path: Path, options: dict) -> None:
    extra = {"brick": components_dir}
    component_options = {**options, **extra}

    create_brick(path, component_options)
    create_test(path, component_options)


def is_brick_dir(p: Path) -> bool:
    return p.is_dir() and p.name not in {"__pycache__", ".venv", ".mypy_cache"}


def get_component_dirs(root: Path, top_dir, ns) -> list:
    theme = configuration.get_theme_from_config(root)
    dirs = top_dir if theme == "tdd" else f"{top_dir}/{ns}"

    component_dir = root / dirs

    if not component_dir.exists():
        return []

    return [f for f in component_dir.iterdir() if is_brick_dir(f)]


def get_components_data(
    root: Path, ns: str, top_dir: str = components_dir
) -> List[dict]:
    dirs = get_component_dirs(root, top_dir, ns)

    return [{"name": d.name} for d in dirs]



from pathlib import Path
from typing import Set

from polylith import configuration, repo


def get_path(structure: str, brick: str, ns: str, package: str) -> str:
    return structure.format(brick=brick, namespace=ns, package=package)


def get_paths(structure: str, brick: str, ns: str, packages: Set[str]) -> Set[str]:
    return {get_path(structure, brick, ns, p) for p in packages}


def collect_paths(root: Path, ns: str, brick: str, packages: Set[str]) -> Set[Path]:
    structure = configuration.get_brick_structure_from_config(root)

    paths = get_paths(structure, brick, ns, packages)

    return {Path(root / p) for p in paths}


def collect_bases_paths(root: Path, ns: str, bases: Set[str]) -> Set[Path]:
    return collect_paths(root, ns, repo.bases_dir, bases)


def collect_components_paths(root: Path, ns: str, components: Set[str]) -> Set[Path]:
    return collect_paths(root, ns, repo.components_dir, components)



import sys
from pathlib import Path
from typing import Set

from polylith import configuration
from polylith.imports import extract_top_ns, fetch_all_imports
from polylith.libs.stdlib import standard_libs


def get_python_version() -> str:
    return f"{sys.version_info.major}.{sys.version_info.minor}"


def get_latest_standard_libs() -> Set[str]:
    values = list(standard_libs.values())

    return values[-1]


def get_standard_libs(python_version: str) -> Set[str]:
    libs = standard_libs.get(python_version)

    return libs or get_latest_standard_libs()


def exclude_libs(import_data: dict, to_exclude: Set) -> dict:
    return {k: v - to_exclude for k, v in import_data.items()}


def exclude_empty(import_data: dict) -> dict:
    return {k: v for k, v in import_data.items() if v}


def extract_third_party_imports(all_imports: dict, top_ns: str) -> dict:
    python_version = get_python_version()
    std_libs = get_standard_libs(python_version)

    top_level_imports = extract_top_ns(all_imports)
    with_third_party = exclude_libs(top_level_imports, std_libs.union({top_ns}))

    return exclude_empty(with_third_party)


def get_third_party_imports(root: Path, paths: Set[Path]) -> dict:
    top_ns = configuration.get_namespace_from_config(root)

    all_imports = fetch_all_imports(paths)

    return extract_third_party_imports(all_imports, top_ns)



import re
import sys
from functools import reduce
from pathlib import Path
from typing import List, Union

import tomlkit
from polylith import repo

if sys.version_info < (3, 11):
    import tomlkit as tomllib
else:
    import tomllib


def transform_to_package(namespace: str, include: str) -> dict:
    path, _separator, brick = str.partition(include, f"/{namespace}/")

    return {"include": f"{namespace}/{brick}", "from": path}


def get_polylith_section(data) -> dict:
    return data.get("tool", {}).get("polylith", {})


def get_custom_top_namespace_from_polylith_section(data) -> Union[str, None]:
    poly_data = get_polylith_section(data)

    return poly_data.get("build", {}).get("top-namespace")


def get_project_packages_from_polylith_section(data) -> dict:
    poly_data = get_polylith_section(data)

    bricks = poly_data.get("bricks")

    return bricks if isinstance(bricks, dict) else {}


def get_hatch_project_packages(data) -> dict:
    polylith_section = get_project_packages_from_polylith_section(data)

    if polylith_section:
        return polylith_section

    hatch_data = data["tool"]["hatch"]
    build_data = hatch_data.get("build", {}) if isinstance(hatch_data, dict) else {}

    return build_data.get("force-include", {})


def get_project_package_includes(namespace: str, data) -> List[dict]:
    if repo.is_poetry(data):
        return data["tool"]["poetry"].get("packages", [])

    includes = (
        get_hatch_project_packages(data)
        if repo.is_hatch(data)
        else get_project_packages_from_polylith_section(data)
    )

    return [transform_to_package(namespace, key) for key in includes.keys()]


def parse_pep_621_dependency(dep: str) -> dict:
    parts = re.split(r"[\^~=!<>]", dep)

    name, *_ = parts if parts else [""]
    version = str.replace(dep, name, "")

    return {name: version} if name else {}


def parse_poetry_dependency(acc: dict, kv: tuple) -> dict:
    k, v = kv

    if isinstance(v, dict):
        extras = sorted(v.get("extras", []))
        version = v.get("version", "")

        name = k + str.replace(f"{extras}", "'", "") if extras else k
        parsed = {name: version}
    else:
        parsed = {k: v}

    return {**acc, **parsed}


def get_pep_621_optional_dependencies(data) -> List[str]:
    groups = data["project"].get("optional-dependencies", {})
    matrix = [v for v in groups.values()] if isinstance(groups, dict) else []

    return sum(matrix, [])


def parse_project_dependencies(data) -> dict:
    if repo.is_poetry(data):
        deps = data["tool"]["poetry"].get("dependencies", {})
        res: dict = reduce(parse_poetry_dependency, deps.items(), {})

        return res

    deps = data["project"].get("dependencies", [])
    optional_deps = get_pep_621_optional_dependencies(data)

    all_deps = deps + optional_deps

    return {k: v for dep in all_deps for k, v in parse_pep_621_dependency(dep).items()}


def get_project_dependencies(data) -> dict:
    items = parse_project_dependencies(data)

    return {"items": items, "source": repo.default_toml}


def read_toml_document(path: Path) -> tomlkit.TOMLDocument:
    with path.open(encoding="utf-8", errors="ignore") as f:
        return tomlkit.loads(f.read())


def load_toml(path: Path) -> dict:
    with open(path, "rb") as f:
        try:
            return tomllib.load(f)
        except tomlkit.exceptions.ParseError as e:
            raise ValueError(f"Failed loading {path}: {repr(e)}") from e



from pathlib import Path

from polylith import info
from polylith import sync


def run(root: Path, ns: str, project_data: dict, options: dict):
    is_quiet = options["quiet"]
    is_verbose = options["verbose"]

    bases = info.get_bases(root, ns)
    components = info.get_components(root, ns)
    workspace_data = {"bases": bases, "components": components}

    diff = sync.calculate_diff(root, ns, project_data, workspace_data)

    sync.update_project(root, ns, diff)

    if is_quiet:
        return

    sync.report.print_summary(diff)

    if is_verbose:
        sync.report.print_brick_imports(diff)



from polylith.toml.core import (
    get_custom_top_namespace_from_polylith_section,
    get_project_dependencies,
    get_project_package_includes,
    get_project_packages_from_polylith_section,
    load_toml,
    parse_project_dependencies,
    read_toml_document,
)

__all__ = [
    "get_custom_top_namespace_from_polylith_section",
    "get_project_dependencies",
    "get_project_package_includes",
    "get_project_packages_from_polylith_section",
    "load_toml",
    "parse_project_dependencies",
    "read_toml_document",
]



from functools import lru_cache
from pathlib import Path

import tomlkit
from polylith.repo.repo import default_toml, is_pep_621_ready


@lru_cache
def get_pyproject_data(path: Path) -> tomlkit.TOMLDocument:
    with open(str(path / default_toml), "r", errors="ignore") as f:
        return tomlkit.loads(f.read())


def get_metadata_section(data: dict) -> dict:
    return data["project"] if is_pep_621_ready(data) else data["tool"]["poetry"]


def get_authors(path: Path) -> str:
    data = get_pyproject_data(path)
    section = get_metadata_section(data)

    return section.get("authors", []).as_string()


def get_python_version(path: Path) -> str:
    data: dict = get_pyproject_data(path)

    if is_pep_621_ready(data):
        return data["project"].get("requires-python", "")

    return data["tool"]["poetry"]["dependencies"]["python"]



from cleo.helpers import option
from poetry.console.commands.command import Command
from polylith import commands


class DiffCommand(Command):
    name = "poly diff"
    description = "Shows changed bricks compared to the latest git tag."

    options = [
        option(
            long_name="short",
            short_name="s",
            description="Print short view",
            flag=True,
        ),
        option(
            long_name="bricks",
            description="Print changed bricks",
            flag=True,
        ),
        option(
            long_name="deps",
            description="Print bricks that depend on the changes. Use with --bricks.",
            flag=True,
        ),
        option(
            long_name="since",
            description="Changed since a specific tag",
            flag=False,
        ),
    ]

    def handle(self) -> int:
        since = self.option("since")

        options = {
            "short": self.option("short"),
            "bricks": self.option("bricks"),
            "deps": self.option("deps"),
        }

        commands.diff.run(since, options)

        return 0



from pathlib import Path

from polylith import configuration
from polylith.dirs import create_dir
from polylith.files import create_file
from polylith.interface import create_interface
from polylith.readme import create_brick_readme


def create_brick(root: Path, options: dict) -> None:
    modulename = options["modulename"]
    path_kwargs = {
        k: v for k, v in options.items() if k in {"brick", "namespace", "package"}
    }

    brick_structure = configuration.get_brick_structure_from_config(root)
    resources_structure = configuration.get_resources_structure_from_config(root)

    brick_path = brick_structure.format(**path_kwargs)
    resources_path = resources_structure.format(**path_kwargs)

    d = create_dir(root, brick_path)
    create_file(d, f"{modulename}.py")
    create_interface(d, options)

    if configuration.is_readme_generation_enabled(root):
        create_brick_readme(root / resources_path, options)



from pathlib import Path

from polylith.dirs import create_dir
from polylith.repo import development_dir


def create_development(path: Path, keep=True) -> None:
    create_dir(path, development_dir, keep=keep)



from functools import partial
from pathlib import Path

from cleo.helpers import option
from poetry.console.commands.command import Command
from polylith import commands, configuration, info, repo
from polylith.poetry import internals

command_options = [
    option(
        long_name="strict",
        description="More strict checks when matching name and version of third-party libraries and imports.",
        flag=True,
    ),
    option(
        long_name="alias",
        description="alias for a third-party library, useful when an import differ from the library name",
        flag=False,
        multiple=True,
    ),
]


class CheckCommand(Command):
    name = "poly check"
    description = "Validates the <comment>Polylith</> workspace."

    options = command_options

    def print_report(
        self, root: Path, ns: str, project_data: dict, options: dict
    ) -> bool:
        path = project_data["path"]
        name = project_data["name"]

        try:
            third_party_libs = internals.find_third_party_libs(self.poetry, path)
            merged = {
                **project_data,
                **{"deps": {"items": third_party_libs, "source": "poetry.lock"}},
            }

            return commands.check.run(root, ns, merged, options)
        except ValueError as e:
            self.line_error(f"{name}: <error>{e}</error>")
            return False

    def handle(self) -> int:
        directory = self.option("directory")
        root = repo.get_workspace_root(Path.cwd())
        ns = configuration.get_namespace_from_config(root)

        all_projects_data = info.get_projects_data(root, ns)
        only_projects_data = [p for p in all_projects_data if info.is_project(p)]

        projects_data = internals.filter_projects_data(
            self.poetry, directory, only_projects_data
        )

        dists_fn = partial(internals.distributions, None, root)
        options = {
            "verbose": True if self.option("verbose") else False,
            "short": False,
            "quiet": True if self.option("quiet") else False,
            "strict": True if self.option("strict") else False,
            "alias": self.option("alias") or [],
            "dists_fn": dists_fn,
        }

        results = {self.print_report(root, ns, data, options) for data in projects_data}

        libs_result = commands.check.check_libs_versions(
            projects_data, all_projects_data, options
        )

        return 0 if all(results) and libs_result else 1



from rich.theme import Theme

poly_theme = Theme(
    {
        "data": "#999966",
        "proj": "#8A2BE2",
        "comp": "#32CD32",
        "base": "#6495ED",
    }
)

check_emoji = ":heavy_check_mark:"



from functools import lru_cache
from pathlib import Path
from typing import List

import tomlkit
from polylith import configuration, repo, toml


def get_project_name(toml_data) -> str:
    if repo.is_pep_621_ready(toml_data):
        return toml_data["project"]["name"]

    return toml_data["tool"]["poetry"]["name"]


def get_project_name_from_toml(data: dict) -> str:
    try:
        return get_project_name(data["toml"])
    except KeyError as e:
        path = data["path"]

        raise KeyError(f"Error in {path}") from e


@lru_cache
def get_toml(path: Path) -> tomlkit.TOMLDocument:
    return toml.read_toml_document(path)


def get_project_files(root: Path) -> dict:
    projects = sorted(root.glob(f"projects/*/{repo.default_toml}"))
    development = Path(root / repo.default_toml)

    proj = {"projects": projects}
    dev = {"development": [development]}

    return {**proj, **dev}


def toml_data(path: Path, project_type: str) -> dict:
    return {"toml": get_toml(path), "path": path.parent, "type": project_type}


def get_toml_files(root: Path) -> List[dict]:
    project_files = get_project_files(root)

    proj = [toml_data(p, "project") for p in project_files["projects"]]
    dev = [toml_data(d, "development") for d in project_files["development"]]

    return proj + dev


def get_packages_for_projects(root: Path) -> List[dict]:
    toml_files = get_toml_files(root)
    namespace = configuration.get_namespace_from_config(root)

    return [
        {
            "name": get_project_name_from_toml(d),
            "packages": toml.get_project_package_includes(namespace, d["toml"]),
            "path": d["path"],
            "type": d["type"],
            "deps": toml.get_project_dependencies(d["toml"]),
        }
        for d in toml_files
    ]



from functools import reduce
from typing import Set

from polylith import alias
from polylith.distributions.core import (
    distributions_packages,
    distributions_sub_packages,
    get_distributions,
    get_packages_distributions,
)


def extract_extras(name: str) -> Set[str]:
    chars = ["[", "]"]
    replacement = ","

    res = reduce(lambda acc, char: str.replace(acc, char, replacement), chars, name)

    parts = str.split(res, replacement)

    return {str.strip(p) for p in parts if p}


def extract_library_names(deps: dict) -> Set[str]:
    names = {k for k, _v in deps["items"].items()}

    with_extras = [extract_extras(n) for n in names]

    return set().union(*with_extras)


def known_aliases_and_sub_dependencies(
    deps: dict, library_alias: list, options: dict
) -> Set[str]:
    """Collect known aliases (packages) for third-party libraries.

    When the library origin is not from a lock-file:
    collect sub-dependencies and distribution top-namespace for each library, and append to the result.
    """

    lock_file = any(str.endswith(deps["source"], s) for s in {".lock", ".txt"})
    third_party_libs = extract_library_names(deps)

    fn = options.get("dists_fn", get_distributions)
    dists = fn()

    dist_packages = distributions_packages(dists)
    custom_aliases = alias.parse(library_alias)
    sub_deps = distributions_sub_packages(dists) if not lock_file else {}

    a = alias.pick(dist_packages, third_party_libs)
    b = alias.pick(custom_aliases, third_party_libs)
    c = alias.pick(sub_deps, third_party_libs)
    d = get_packages_distributions(third_party_libs)

    return third_party_libs.union(a, b, c, d)



from functools import lru_cache
from pathlib import Path
from typing import List, Union

import tomlkit
from polylith import repo


@lru_cache
def _load_workspace_config(path: Path) -> tomlkit.TOMLDocument:
    fullpath = path / repo.workspace_file

    if not fullpath.exists():
        fullpath = path / repo.default_toml

    content = fullpath.read_text()

    return tomlkit.loads(content)


def get_namespace_from_config(path: Path) -> str:
    toml: dict = _load_workspace_config(path)

    return toml["tool"]["polylith"]["namespace"]


def get_git_tag_pattern(toml: dict) -> str:
    """Fallback git tag pattern configuration"""
    return toml["tool"]["polylith"]["git_tag_pattern"]


def get_tag_pattern_from_config(path: Path, key: Union[str, None]) -> str:
    toml: dict = _load_workspace_config(path)

    patterns = toml["tool"]["polylith"].get("tag", {}).get("patterns")

    return patterns[key or "stable"] if patterns else get_git_tag_pattern(toml)


def get_tag_sort_options_from_config(path: Path) -> List[str]:
    toml: dict = _load_workspace_config(path)

    options = toml["tool"]["polylith"].get("tag", {}).get("sorting")
    # Default sorting option
    if options is None:
        return ["-committerdate"]
    return options


def is_test_generation_enabled(path: Path) -> bool:
    toml: dict = _load_workspace_config(path)

    enabled = toml["tool"]["polylith"]["test"]["enabled"]
    return bool(enabled)


def is_readme_generation_enabled(path: Path) -> bool:
    toml: dict = _load_workspace_config(path)

    enabled = toml["tool"]["polylith"].get("resources", {}).get("brick_docs_enabled")
    return bool(enabled)


def get_theme_from_config(path: Path) -> str:
    toml: dict = _load_workspace_config(path)

    return toml["tool"]["polylith"]["structure"].get("theme") or "tdd"


def get_brick_structure_from_config(path: Path) -> str:
    theme = get_theme_from_config(path)

    if theme == "loose":
        return "{brick}/{namespace}/{package}"

    return "{brick}/{package}/src/{namespace}/{package}"


def get_tests_structure_from_config(path: Path) -> str:
    theme = get_theme_from_config(path)

    if theme == "loose":
        return "test/{brick}/{namespace}/{package}"

    return "{brick}/{package}/test/{namespace}/{package}"


def get_resources_structure_from_config(path: Path) -> str:
    theme = get_theme_from_config(path)

    if theme == "loose":
        return "{brick}/{namespace}/{package}"

    return "{brick}/{package}"



from cleo.helpers import option
from poetry.console.commands.command import Command
from polylith import commands


class InfoCommand(Command):
    name = "poly info"
    description = "Info about the <comment>Polylith</> workspace."

    options = [
        option(
            long_name="short",
            short_name="s",
            description="Display Workspace Info adjusted for many projects",
            flag=True,
        ),
    ]

    def handle(self) -> int:
        short = True if self.option("short") else False

        commands.info.run(short)

        return 0



import re
import subprocess
from pathlib import Path
from typing import List, Set, Union

from polylith import configuration, repo


def _parse_folder_parts(pattern: str, changed_file: Path) -> str:
    parts = re.split(pattern, changed_file.as_posix())
    remainder = parts[-1]

    file_path = Path(remainder)

    return next(p for p in file_path.parts if p != file_path.root)


def _get_changed(pattern: str, changed_files: List[Path]) -> set:
    return {
        _parse_folder_parts(pattern, f)
        for f in changed_files
        if re.match(pattern, f.as_posix())
    }


def _parse_path_pattern(root: Path, top_dir: str, namespace: str) -> str:
    theme = configuration.get_theme_from_config(root)

    if theme == "loose":
        return f"{top_dir}/{namespace}/"

    return rf"({top_dir}\/).+(\/src\/)({namespace})"


def _get_changed_bricks(
    root: Path, top_dir: str, changed_files: List[Path], namespace: str
) -> list:
    pattern = _parse_path_pattern(root, top_dir, namespace)

    return sorted(_get_changed(pattern, changed_files))


def get_changed_components(
    root: Path, changed_files: List[Path], namespace: str
) -> list:
    return _get_changed_bricks(root, repo.components_dir, changed_files, namespace)


def get_changed_bases(root: Path, changed_files: List[Path], namespace: str) -> list:
    return _get_changed_bricks(root, repo.bases_dir, changed_files, namespace)


def get_changed_projects(changed_files: List[Path]) -> list:
    res = _get_changed(repo.projects_dir, changed_files)
    filtered = {p for p in res if p != repo.projects_dir}
    return sorted(filtered)


def get_latest_tag(root: Path, key: Union[str, None]) -> Union[str, None]:
    tag_pattern = configuration.get_tag_pattern_from_config(root, key)
    sorting_options = [
        f"--sort={option}"
        for option in configuration.get_tag_sort_options_from_config(root)
    ]

    res = subprocess.run(
        ["git", "tag", "-l"] + sorting_options + [f"{tag_pattern}"],
        capture_output=True,
    )

    return next((tag for tag in res.stdout.decode("utf-8").split()), None)


def get_files(tag: str) -> List[Path]:
    res = subprocess.run(
        ["git", "diff", tag, "--stat", "--name-only"],
        capture_output=True,
    )

    return [Path(p) for p in res.stdout.decode("utf-8").split()]


def _affected(projects_data: List[dict], brick_type: str, bricks: List[str]) -> set:
    res = {
        p["path"].name: set(p.get(brick_type, [])).intersection(bricks)
        for p in projects_data
    }

    return {k for k, v in res.items() if v}


def get_projects_affected_by_changes(
    projects_data: List[dict],
    projects: List[str],
    bases: List[str],
    components: List[str],
) -> Set[str]:
    a = _affected(projects_data, "components", components)
    b = _affected(projects_data, "bases", bases)
    c = set(projects)

    return {*a, *b, *c}



from pathlib import Path

import tomlkit
from polylith import repo
from polylith.dirs import create_dir
from polylith.repo import projects_dir


def create_project_toml(template: str, template_data: dict) -> tomlkit.TOMLDocument:
    content = template.format(**template_data)

    return tomlkit.loads(content)


def create_project(path: Path, template: str, name: str, description: str) -> None:
    d = create_dir(path, f"{projects_dir}/{name}")

    authors = repo.get_authors(path)
    python_version = repo.get_python_version(path)

    project_toml = create_project_toml(
        template,
        {
            "name": name,
            "description": description,
            "authors": authors,
            "python_version": python_version,
        },
    )

    fullpath = d / repo.default_toml

    with fullpath.open("w", encoding="utf-8") as f:
        f.write(tomlkit.dumps(project_toml))



from pathlib import Path
from typing import Union

workspace_file = "workspace.toml"
root_file = ".git"
default_toml = "pyproject.toml"
readme_file = "README.md"

bases_dir = "bases"
components_dir = "components"
projects_dir = "projects"
development_dir = "development"


def is_drive_root(cwd: Path) -> bool:
    return cwd == Path(cwd.root) or cwd == cwd.parent


def is_repo_root(cwd: Path) -> bool:
    fullpath = cwd / root_file

    return fullpath.exists()


def find_upwards(cwd: Path, name: str) -> Union[Path, None]:
    if is_drive_root(cwd):
        return None

    fullpath = cwd / name

    if fullpath.exists():
        return fullpath

    if is_repo_root(cwd):
        return None

    return find_upwards(cwd.parent, name)


def find_upwards_dir(cwd: Path, name: str) -> Union[Path, None]:
    fullpath = find_upwards(cwd, name)

    return fullpath.parent if fullpath else None


def find_workspace_root(cwd: Path) -> Union[Path, None]:
    workspace_root = find_upwards_dir(cwd, workspace_file)
    if workspace_root:
        return workspace_root
    return find_upwards_dir(cwd, root_file)


def get_workspace_root(cwd: Path) -> Path:
    root = find_workspace_root(cwd)

    if not root:
        raise ValueError(
            "Didn't find the workspace root. Expected to find a workspace.toml or .git file."
        )

    return root


def has_build_requires(pyproject: dict, value: str) -> bool:
    backend = pyproject.get("build-system", {}).get("build-backend", "")

    return value in backend


def is_poetry(pyproject: dict) -> bool:
    return has_build_requires(pyproject, "poetry")


def is_hatch(pyproject: dict) -> bool:
    return has_build_requires(pyproject, "hatchling")


def is_pdm(pyproject: dict) -> bool:
    return has_build_requires(pyproject, "pdm")


def is_pep_621_ready(pyproject: dict) -> bool:
    if is_poetry(pyproject):
        return False

    return pyproject.get("project", {}).get("name") is not None



from functools import reduce
from pathlib import Path
from typing import List, Union

import tomlkit
from polylith import configuration, project, repo
from tomlkit.toml_document import TOMLDocument


def to_package(namespace: str, brick: str, brick_path: str, theme: str) -> dict:
    folder = f"{brick_path}" if theme == "loose" else f"{brick_path}/{brick}/src"

    return {"include": f"{namespace}/{brick}", "from": folder}


def copy_toml_data(data: TOMLDocument) -> dict:
    original = tomlkit.dumps(data)
    copy: dict = tomlkit.parse(original)

    return copy


def to_key_value_include(acc: dict, package: dict) -> dict:
    brick = package["include"]
    relative_path = package.get("from", "")
    include = Path(relative_path, brick).as_posix()

    return {**acc, **{include: brick}}


def generate_updated_pep_621_project(data: TOMLDocument, bricks_to_add: dict) -> str:
    copy = copy_toml_data(data)

    if not copy.get("tool"):
        copy["tool"] = {"polylith": {"bricks": {}}}

    if not copy["tool"].get("polylith"):
        copy["tool"]["polylith"] = {"bricks": {}}

    if not copy["tool"]["polylith"].get("bricks"):
        copy["tool"]["polylith"]["bricks"] = {}

    for k, v in bricks_to_add.items():
        copy["tool"]["polylith"]["bricks"][k] = v

    return tomlkit.dumps(copy)


def generate_updated_pdm_project(data: TOMLDocument, packages: List[dict]) -> str:
    bricks_to_add: dict = reduce(to_key_value_include, packages, {})

    return generate_updated_pep_621_project(data, bricks_to_add)


def generate_updated_hatch_project(data: TOMLDocument, packages: List[dict]) -> str:
    bricks_to_add: dict = reduce(to_key_value_include, packages, {})

    has_polylith = data.get("tool", {}).get("polylith", {}).get("bricks")
    has_hatch = (
        data.get("tool", {}).get("hatch", {}).get("build", {}).get("force-include")
    )

    if not has_polylith and has_hatch:
        copy = copy_toml_data(data)

        for k, v in bricks_to_add.items():
            copy["tool"]["hatch"]["build"]["force-include"][k] = v

        return tomlkit.dumps(copy)

    return generate_updated_pep_621_project(data, bricks_to_add)


def generate_updated_poetry_project(data: TOMLDocument, packages: List[dict]) -> str:
    copy = copy_toml_data(data)

    if copy["tool"]["poetry"].get("packages") is None:
        copy["tool"]["poetry"].add("packages", [])

    for package in packages:
        copy["tool"]["poetry"]["packages"].append(package)

    copy["tool"]["poetry"]["packages"].multiline(True)

    return tomlkit.dumps(copy)


def generate_updated_project(
    data: TOMLDocument, packages: List[dict]
) -> Union[str, None]:
    if repo.is_poetry(data):
        return generate_updated_poetry_project(data, packages)

    if repo.is_hatch(data):
        return generate_updated_hatch_project(data, packages)

    if repo.is_pdm(data):
        return generate_updated_pdm_project(data, packages)

    return None


def to_packages(root: Path, namespace: str, diff: dict) -> List[dict]:
    theme = configuration.get_theme_from_config(root)

    is_project = diff["is_project"]

    bases_path = "../../bases" if is_project else "bases"
    components_path = "../../components" if is_project else "components"

    a = [to_package(namespace, b, bases_path, theme) for b in diff["bases"]]
    b = [to_package(namespace, c, components_path, theme) for c in diff["components"]]

    return a + b


def rewrite_project_file(path: Path, packages: List[dict]):
    fullpath = path / repo.default_toml
    project_toml = project.get_toml(fullpath)

    generated = generate_updated_project(project_toml, packages)

    if not generated:
        return

    with fullpath.open("w", encoding="utf-8") as f:
        f.write(generated)


def update_project(path: Path, namespace: str, diff: dict):
    packages = to_packages(path, namespace, diff)

    if packages:
        rewrite_project_file(diff["path"], packages)



from pathlib import Path

from polylith.dirs import create_dir
from polylith.files import create_file
from polylith import configuration

template = """\
from {namespace}.{package} import {modulename}


def test_sample():
    assert {modulename} is not None
"""


def create_test(root: Path, options: dict) -> None:
    if not configuration.is_test_generation_enabled(root):
        return

    brick = options["brick"]
    namespace = options["namespace"]
    package = options["package"]
    modulename = options["modulename"]

    dirs_structure = configuration.get_tests_structure_from_config(root)
    dirs = dirs_structure.format(brick=brick, namespace=namespace, package=package)
    d = create_dir(root, dirs)

    create_file(d, "__init__.py")
    test_file = create_file(d, f"test_{modulename}.py")

    content = template.format(
        namespace=namespace, package=package, modulename=modulename
    )

    test_file.write_text(content)
